#if 0 /*
# -----------------------------------------------------------------------
# imap.py - IMAP4 protocol implementation
# Author: Viggo Fredriksen <viggo@katatonic.org>
# -----------------------------------------------------------------------
# $Id: imap.py,v 1.13 2004/05/19 10:35:04 viggo Exp $
# -----------------------------------------------------------------------
# $Log: imap.py,v $
# Revision 1.13  2004/05/19 10:35:04  viggo
# Fixed imap expunge() handling
#
#
# -----------------------------------------------------------------------
# Todo:
#  - implement more imap features?
#  - Only fetch the XXX last mails?
# -----------------------------------------------------------------------
# Freevo - A Home Theater PC framework
# Copyright (C) 2003 Krister Lagerstrom, et al.
# Please see the file freevo/Docs/CREDITS for a complete list of authors.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MER-
# CHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#
# ----------------------------------------------------------------------- */
#endif

# python modules
import imaplib, re

# mail modules
from mail import MailProtocol
from mailutil import *



class Imap(MailProtocol):
    """
    Imap implementation
    Need too look closer at IMAP rfc's
    doc: http://www.python.org/doc/current/lib/module-imaplib.html
         http://www.cis.ohio-state.edu/cgi-bin/rfc/rfc2683.html
         http://asg.web.cmu.edu/cyrus/rfc/draft-ietf-imap-disc-01.html
         http://www.cis.ohio-state.edu/cgi-bin/rfc/rfc2060.html
    """
    def __init__(self):
        MailProtocol.__init__(self)
        self.sSort = False
        self.setup = False
        self.lastSeen = {}
        self.flaglist = {}
        self.caps = ['DEL','FLAG', 'SUBS']

        if self.SSL:
            import sys,socket
            if (float(sys.version[0:3]) >= 2.3):
                if socket.getdefaulttimeout():
                    _debug_('Changing default socket timout (because of pymetar).')
                    print 'IMAP-SSL:  Removing socket-timeout set in pymetar.py. SSL won\'t work with it'
                    socket.setdefaulttimeout(None)

    def _getServerCap(self):
        """
        parse necessery server capabilities here
        """
        self.setup = True
        self.separator = '/'

        # look for serverside sort
        if 'SORT' in self.M.capabilities:
            self.sSort = True

        # get separator
        try:
            s =  self.M.list()
            if s[0] == 'OK':
                self.separator = re.match('\(.*\) \"(.+)\" .*', s[1][0]).groups()[0]
        except IndexError:
            _debug_('Failed to retrieve IMAP hierarchy separator, trying default')

    def _FolderFormat(self,data):
        """
        format folders
        Examples:
           Exchange
            ['() "/" INBOX', '() "/" "Sent Items"', ....]
           Courier:
            ['(\\HasNoChildren) "." "INBOX.Trash"', ....]
        This could be used to add hierachi to the
        folders in future.
        """
        foldernames = []

        for f in data:
            reg = re.match('\(.*\) \".+\" (.*)' , f)
            val = reg.groups()[0].replace('"','')
            foldernames += [ val ]

        return foldernames

    def _getAllFolders(self):
        """
        get all folders on server
        """
        typ,data = self.M.list()

        if typ == 'OK':
            return self._FolderFormat(data)
        else:
            return [ 'INBOX' ]

    def _select(self,folder,readonly=1):
        try:
            val = self.M.select(folder,readonly)
            if val[0] == 'OK': return True
        except imaplib.IMAP4.error:
            _debug_('Error selecting imap folder')
        return False

    def _parseHeader(self, header, flags, folder):
        header = getHeader(header)
        header.UID = flags.split(' ')[2]
        lst = int(header.UID)
        if lst>self.lastSeen[folder]:
            self.lastSeen[folder] = lst
        header.Flag = self._parseFlags(flags)
        return header

    def _parseFlags(self,Flags):
        flags = imaplib.ParseFlags(Flags)

        if '\\Seen' in flags:
            return 'seen'
        elif '\\Answered' in flags:
            return 'seen'
        elif '\\UnSeen' in flags:
            return 'new'
        else:
            return 'new'

    def login(self):
        try:
            if self.SSL:
                self.M = imaplib.IMAP4_SSL(self.host)
            else:
                self.M = imaplib.IMAP4(self.host)

            self.M.login(self.username, self.password)
            if not self.setup: self._getServerCap()
            return 'OK'
        except Exception, err:
            return '%s' % err

    def logout(self):
        try:
            self.M.logout()
        except imaplib.IMAP4.error:
            _debug_('Mail: Something went wrong when closing IMAP connection')

    def getFolders(self):
        folders = []

        typ,data = self.M.lsub()
        if typ == 'OK' and data[0]:
            folders = self._FolderFormat(data)

        return folders

    def getUnsubscribed(self):
        """
        get folders we are not subscribed to
        """
        folders = []

        us_name = self._getAllFolders()
        s_name = self.getFolders()

        for us in us_name:
            if us not in s_name:
                folders.append(us)

        return folders

    def delMessage(self, header):
        success = False
        if self._select(header.Folder,readonly=None):
            if self.M.uid('STORE', header.UID, '+FLAGS', '(\Deleted)')[0] == 'OK':
                if self.M.expunge()[0] == 'OK':
                    success = True
            self.M.close()
        return success

    def flagMessage(self, header, action='READ'):
        """
        flag a message
        """
        flag = '+FLAGS'
        if action == 'UNREAD':
            flag = '-FLAGS'

        if self._select(header.Folder, readonly=None):
            self.M.uid('STORE', header.UID, flag, '(\Seen)')
            self.M.close()
            return True

    def markAllRead(self, folder=None):
        if self._select(folder, readonly=None):
            self.M.uid('STORE', '1:*', '+FLAGS', '(\Seen)')
            self.M.close()
            return True


    def unsubscribe(self,folder):
        """
        unsubscribe from a folder
        """
        if self.M.unsubscribe(folder)[0] == 'OK':
            return True

    def subscribe(self,folder):
        """
        subscribe to a folder
        """
        if self.M.subscribe(folder)[0] == 'OK':
            return True

    def getMessage(self,header):
        mail = BaseMessage(header)
        if self._select(header.Folder):
            typ,data = self.M.uid('FETCH',header.UID,'(RFC822)' )
            self.M.close()

            if typ == 'OK' and data[0] != None:
                self.flagMessage(header,action='READ')
                mail = getMessage(data[0][1],header)
        return mail

    def inspectOldHeaders(self,folder,newflags):
        """
        inspects the headers and returns the changed
        uids,flags that has changed + a list of uids
        that should be removed from cache
        """
        upd = []
        rem = []
        hask = newflags.has_key

        for (uid,flag) in self.flaglist[folder]:
            if not hask(uid):
                rem.append(uid)
            else:
                f = newflags[uid]
                if f != flag:
                    upd.append( (uid,uid,f) )
        return upd,rem

    def getUpdates(self, folder, seen):
        """
        Seems like this is not working properly atm
        """
        upd  = []
        rem  = []
        add  = []
        uidl = []

        if self._select(folder):
            # Look for new messages since last visit
            fuids = '%i:*' % (self.lastSeen[folder] + 1)
            last = self.lastSeen[folder]
            typ,data = self.M.uid('FETCH',fuids,'(UID FLAGS RFC822.HEADER)' )
            if typ == 'OK':
                for m in data:
                    if m and len(m)==2:
                        header = self._parseHeader(m[1],m[0],folder)
                        header.Folder = folder
                        uidl.append(header.UID)
                        add.append(header)
                add.reverse()

            # look for updates and removals
            typ,data = self.M.uid('FETCH',('1:%i'%last), '(UID FLAGS)' )
            if typ == 'OK':
                newfd = {}
                newf = []
                app = newf.append
                for d in data:
                    if d:
                        uid = re.match('.*\(UID (.*) FLAGS.*',d).groups()[0]
                        uidl.append(uid)
                        f = self._parseFlags(d)
                        newfd[uid] = f
                        app((uid,f))


                upd,rem = self.inspectOldHeaders(folder,newfd)

                self.flaglist[folder] = newf
            self.M.close()
        return upd,rem,add,uidl

    def getHeaders(self,folder):
        """
        Update this to use uid fetch 1:*
        (if this is possible with SORT)
        """
        mails = []
        flags = []

        if not self.lastSeen.has_key(folder):
            self.lastSeen[folder] = 1

        if self._select(folder):
            if self.sSort:
                typ,data = self.M.sort('(ARRIVAL)', 'US-ASCII', 'ALL' )
                sorted = data[0].split()
                sorted.reverse() # seems to be in reverse on my server
            else:
                typ,data = self.M.search(None, '(ALL UNDELETED)')
                sorted = data[0].split()

            if typ == 'OK':
                for num in sorted:
                    typ,mlist = self.M.fetch(num,'(UID FLAGS RFC822.HEADER)')

                    if typ == 'OK':
                        header = self._parseHeader(mlist[0][1], mlist[0][0], folder)
                        header.Folder = folder
                        mails.append(header)
                        flags.append( (header.UID,header.Flag) )

            self.M.close()
            # update internal cache
            self.flaglist[folder] = flags

        if len(mails)>0 and not self.sSort:
            return self.Sort(mails)
        else:
            return mails

    def getHeader(self, folder, uid):
        header = BaseHeader()
        if self._select(folder):
            typ,data = self.M.uid('FETCH',uid,'(UID FLAGS RFC822.HEADER)' )
            self.M.close()
            if typ == 'OK':
                header = self._parseHeader(data[0][1],data[0][0], folder)
                header.Folder = folder
        return header

    def purgeFolder(self, folder):
        if self._select(folder,readonly=None):
            typ, msgs = self.M.search(None, '(ALL UNDELETED)')
            if typ == 'OK':
                for num in msgs[0].split(' '):
                    self.M.store(num, '+FLAGS', '(\Deleted)')
                self.M.expunge()
